from typing import Tuple
import torch
from torch import nn

def _waveform_check(x: torch.Tensor):
  assert x.ndim == 2, "input must have two dimensions (channels, samples)"
  assert x.shape[-1] > x.shape[0], f"The number of channels {x.shape[-2]} exceeds the number of samples {x.shape[-1]} in your INPUT waveform. \
                                      There might be something wrong with your model. "

class AudacityModel(nn.Module):

  def __init__(self, model: nn.Module):
    """ creates an Audacity model, wrapping a child model (that does the real work)"""
    super().__init__()
    model.eval()
    self.model = model

class WaveformToWaveformBase(AudacityModel):

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    """ 
    Internal forward pass for a WaveformToWaveform model. 

    All this does is wrap the do_forward_pass(x) function in assertions that check 
    that the correct input/output constraints are getting met. Nothing fancy. 
    """
    _waveform_check(x)
    x = self.do_forward_pass(x)
    _waveform_check(x)
    
    return x

  def do_forward_pass(self, x: torch.Tensor) -> torch.Tensor:
    """ 
    Perform a forward pass on a waveform-to-waveform model.
    
    Args:
        x : An input audio waveform tensor. If `"multichannel" == True` in the 
            model's `metadata.json`, then this tensor will always be shape 
            `(1, n_samples)`, as all incoming audio will be downmixed first. 
            Otherwise, expect `x` to be a multichannel waveform tensor with 
            shape `(n_channels, n_samples)`.

    Returns:
        torch.Tensor: Output tensor, shape (n_sources, n_samples). Each source 
                      will be  
    """
    raise NotImplementedError("implement me!")

class WaveformToLabelsBase(AudacityModel):

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    """ 
    Internal forward pass for a WaveformToLabels model. 

    All this does is wrap the do_forward_pass(x) function in assertions that check 
    that the correct input/output constraints are getting met. Nothing fancy. 
    """
    _waveform_check(x)
    output = self.do_forward_pass(x)

    assert isinstance(output, tuple), "waveform-to-labels output must be a tuple"
    assert len(output) == 2, "output tuple must have two elements, e.g. tuple(labels, timestamps)"

    labels = output[0]
    timestamps = output[1]

    assert labels.shape[0] == timestamps.shape[0], "time dimension between "\
                                    "labels and timestamps tensors must be equal"
    assert timestamps.shape[1] == 2, "second dimension of the timestamps tensor"\
                                      "must be size 2"

  def do_forward_pass(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """ 
    Perform a forward pass on a waveform-to-labels model.
    
    Args:
        x : An input audio waveform tensor. If `"multichannel" == True` in the 
            model's `metadata.json`, then this tensor will always be shape 
            `(1, n_samples)`, as all incoming audio will be downmixed first. 
            Otherwise, expect `x` to be a multichannel waveform tensor with 
            shape `(n_channels, n_samples)`.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: a tuple of tensors, where the first 
            tensor contains the output class probabilities 
            (shape `(n_timesteps, n_labels)`), and the second tensor contains 
            timestamps with start and end times for each label, 
            shape `(n_timesteps, 2)`. 

    """
    raise NotImplementedError("implement me!")



from pathlib import Path
import random
from typing import Tuple
from audacitorch import AudacityModel

import torch
import json

def save_model(model: torch.jit.ScriptModule, metadata: dict, root_dir: Path):
  """
  Save a compiled torch.jit.ScriptModule, along with a metadata dictionary.

  Args:
      model: your Audacity-ready serialized model, using either torch.jit.trace or torch.jit.script. 
        Should derive from audacitorch.WaveformToWaveformBase or audacitorch.WaveformToLabelsBase.
      metadata: a metadata dictionary. Shoule be validated using torchaudio.utils.validate_metadata()

  Returns:
    Will create the following files: 
    ```
      root_dir/
      root_dir/model.pt
      root_dir/metadata.json
    ```
  """
  root_dir.mkdir(exist_ok=True, parents=True)

  # save model and metadata!
  torch.jit.save(model, root_dir / 'model.pt')

  with open(root_dir / 'metadata.json', 'w') as f:
    json.dump(metadata, f)

def get_example_inputs(multichannel: bool = False):
  """
  returns a list of possible input tensors for an AudacityModel. 

  Possible inputs are audio tensors with shape (n_channels, n_samples). 
  If multichannel == False, n_channels will always be 1. 
  """
  max_channels = 2 if multichannel else 1
  num_inputs = 10
  channels = [random.randint(1, max_channels) for _ in range(num_inputs)]
  sizes = [random.randint(2048, 396000) for _ in range(num_inputs)]
  return [
    torch.randn((c, s)) for c, s in  zip(channels, sizes)
  ]

def test_run(model: AudacityModel, multichannel: bool = False):
  """ 
  Performs a couple of test forward passes with audio tensors of different sizes.
  Possible inputs are audio tensors with shape (n_channels, n_samples). 
  If the model fails to meet the input/output requirements of either WaveformToWaveformBase or WaveformToLabelsBase, 
    an assertion will be triggered by the respective class. 

    Args:
      model (AudacityModel): Your model, wrapped in either WaveformToWaveformBase or WaveformToLabelsBase
      multichannel (bool): if False, the number of input audio channels will always equal to 1. Otherwise, 
                           some stereo test input arrays will be generated.  
  Returns:
    
  """
  for x in get_example_inputs(multichannel):
    model(x)

def load_schema():
    """loads the audacity deep learning json schema for metadata"""
    from urllib.request import urlopen

    url = 'https://raw.githubusercontent.com/hugofloresgarcia/audacity/deeplearning/deeplearning-models/modelcard-schema.json'

    response = urlopen(url)

    schema = json.loads(response.read())

    return schema

def validate_metadata(metadata: dict) -> Tuple[bool, str]:
  """validate a model metadata dict using Audacity's metadata schema

  Args:
      metadata (dict): the metadata dictionary to validate

  Returns:
      Tuple[bool, str], where the  bool indicates success, and 
      the string contains an error/success message
  """
  import jsonschema
  from jsonschema import validate
  schema = load_schema()

  try:
    validate(instance=metadata, schema=schema)
  except jsonschema.exceptions.ValidationError as err:
    print(err)
    return False, err

  message = "success! :)"
  return True, message


from setuptools import setup


with open('README.md') as file:
    long_description = file.read()


setup(
    name='audacitorch',
    description='',
    version='0.0.1',
    author='Hugo Flores Garcia',
    author_email='hf01049@georgiasouthern.edu',
    url='https://github.com/hugofloresgarcia/audacitorch',
    install_requires=[
        'torch==1.9.0',
        'jsonschema'
    ],
    packages=['audacitorch'],
    package_data={'audacitorch': ['assets/*']},
    include_package_data=False,
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=[],
    classifiers=['License :: OSI Approved :: MIT License'],
    license='MIT')
